@@ -70,12 +70,15 @@ module Network.Wai.Handler.Warp (
   , setSlowlorisSize
   , setHTTP2Disabled
   , setLogger
+  , setServerPushLogger
+  , setGracefulShutdownTimeout
     -- ** Getters
   , getPort
   , getHost
   , getOnOpen
   , getOnClose
   , getOnException
+  , getGracefulShutdownTimeout
     -- ** Exception handler
   , defaultOnException
   , defaultShouldDisplayException
@@ -91,10 +94,26 @@ module Network.Wai.Handler.Warp (
   , FileInfo(..)
   , getFileInfo
   , withApplication
+  , withApplicationSettings
   , testWithApplication
+  , testWithApplicationSettings
   , openFreePort
     -- * Version
   , warpVersion
+    -- * HTTP/2
+    -- ** HTTP2 data
+  , HTTP2Data
+  , http2dataPushPromise
+  , defaultHTTP2Data
+  , getHTTP2Data
+  , setHTTP2Data
+    -- ** Push promise
+  , PushPromise
+  , promisedPath
+  , promisedFile
+  , promisedResponseHeaders
+  , promisedWeight
+  , defaultPushPromise
   ) where
 
 import Control.Exception (SomeException, throwIO)
@@ -110,6 +129,8 @@ import Network.Wai.Handler.Warp.Request
 import Network.Wai.Handler.Warp.Response (warpVersion)
 import Network.Wai.Handler.Warp.Run
 import Network.Wai.Handler.Warp.Settings
+import Network.Wai.Handler.Warp.HTTP2.Request (getHTTP2Data, setHTTP2Data)
+import Network.Wai.Handler.Warp.HTTP2.Types
 import Network.Wai.Handler.Warp.Timeout
 import Network.Wai.Handler.Warp.Types hiding (getFileInfo)
 import Network.Wai.Handler.Warp.WithApplication
@@ -134,9 +155,18 @@ setOnException :: (Maybe Request -> SomeException -> IO ()) -> Settings -> Setti
 setOnException x y = y { settingsOnException = x }
 
 -- | A function to create a `Response` when an exception occurs.
---
 -- Default: 'defaultOnExceptionResponse'
 --
+-- Note that an application can handle its own exceptions without interfering with Warp:
+--
+-- > myApp :: Application
+-- > myApp request respond = innerApp `catch` onError
+-- >   where
+-- >     onError = respond . response500 request
+-- >
+-- > response500 :: Request -> SomeException -> Response
+-- > response500 req someEx = responseLBS status500 -- ...
+--
 -- Since 2.1.0
 setOnExceptionResponse :: (SomeException -> Response) -> Settings -> Settings
 setOnExceptionResponse x y = y { settingsOnExceptionResponse = x }
@@ -242,6 +272,12 @@ getOnClose = settingsOnClose
 getOnException :: Settings -> Maybe Request -> SomeException -> IO ()
 getOnException = settingsOnException
 
+-- | Get the graceful shutdown timeout
+--
+-- Since 3.2.8
+getGracefulShutdownTimeout :: Settings -> Maybe Int
+getGracefulShutdownTimeout = settingsGracefulShutdownTimeout
+
 -- | A code to install shutdown handler.
 --
 -- For instance, this code should set up a UNIX signal
@@ -338,13 +374,31 @@ setSlowlorisSize x y = y { settingsSlowlorisSize = x }
 setHTTP2Disabled :: Settings -> Settings
 setHTTP2Disabled y = y { settingsHTTP2Enabled = False }
 
--- | Setting a log function. `Integer` is the body length of a response.
+-- | Setting a log function.
 --
 -- Since 3.X.X
-setLogger :: (Request -> H.Status -> Maybe Integer -> IO ())
-          -> Settings -> Settings
+setLogger :: (Request -> H.Status -> Maybe Integer -> IO ()) -- ^ request, status, maybe file-size
+          -> Settings
+          -> Settings
 setLogger lgr y = y { settingsLogger = lgr }
 
+-- | Setting a log function for HTTP/2 server push.
+--
+--   Since: 3.2.7
+setServerPushLogger :: (Request -> ByteString -> Integer -> IO ()) -- ^ request, path, file-size
+                    -> Settings
+                    -> Settings
+setServerPushLogger lgr y = y { settingsServerPushLogger = lgr }
+
+-- | Set the graceful shutdown timeout. A timeout of `Nothing' will
+-- wait indefinitely, and a number, if provided, will be treated as seconds
+-- to wait for requests to finish, before shutting down the server entirely.
+--
+-- Since 3.2.8
+setGracefulShutdownTimeout :: Maybe Int
+                           -> Settings -> Settings
+setGracefulShutdownTimeout time y = y { settingsGracefulShutdownTimeout = time }
+
 -- | Explicitly pause the slowloris timeout.
 --
 -- This is useful for cases where you partially consume a request body. For
