@@ -6,6 +6,7 @@ module Network.Wai.Handler.Warp.HTTP2.HPACK (
   , hpackEncodeHeaderLoop
   , hpackDecodeHeader
   , just
+  , addNecessaryHeaders
   ) where
 
 import qualified Control.Exception as E
@@ -13,7 +14,6 @@ import Control.Monad (unless)
 import Data.ByteString (ByteString)
 import Network.HPACK hiding (Buffer)
 import Network.HPACK.Token
-import qualified Network.HTTP.Types as H
 import Network.HTTP2
 import Network.Wai.Handler.Warp.HTTP2.Types
 import Network.Wai.Handler.Warp.PackInt
@@ -23,23 +23,7 @@ import Network.Wai.Handler.Warp.Types
 -- $setup
 -- >>> :set -XOverloadedStrings
 
-strategy :: EncodeStrategy
-strategy = EncodeStrategy { compressionAlgo = Linear, useHuffman = False }
-
--- Set-Cookie: contains only one cookie value.
--- So, we don't need to split it.
-hpackEncodeHeader :: Context -> Buffer -> BufSize
-                  -> InternalInfo -> S.Settings
-                  -> H.Status -> (TokenHeaderList,ValueTable)
-                  -> IO (TokenHeaderList, Int)
-hpackEncodeHeader Context{..} buf siz ii settings s (ths0,tbl) = do
-    let !defServer = S.settingsServerName settings
-        !ths1 = addHeader tokenServer defServer tbl ths0
-    date <- getDate ii
-    let !ths2 = addHeader tokenDate date tbl ths1
-        !status = packStatus s
-        !ths3 = (tokenStatus, status) : ths2
-    encodeTokenHeader buf siz strategy True encodeDynamicTable ths3
+----------------------------------------------------------------
 
 {-# INLINE addHeader #-}
 addHeader :: Token -> ByteString -> ValueTable -> TokenHeaderList -> TokenHeaderList
@@ -47,7 +31,37 @@ addHeader t v tbl ths = case getHeaderValue t tbl of
     Nothing -> (t,v) : ths
     Just _  -> ths
 
-hpackEncodeHeaderLoop :: Context -> Buffer -> BufSize -> TokenHeaderList
+addNecessaryHeaders :: Context
+                    -> Rspn
+                    -> InternalInfo
+                    -> S.Settings
+                    -> IO TokenHeaderList
+addNecessaryHeaders Context{..} rspn ii settings = do
+    date <- getDate ii
+    let !s = rspnStatus rspn
+        !status = packStatus s
+        !defServer = S.settingsServerName settings
+        (!ths0,tbl) = rspnHeaders rspn
+        !ths1 = addHeader tokenServer defServer tbl ths0
+        !ths2 = addHeader tokenDate date tbl ths1
+        !ths3 = (tokenStatus, status) : ths2
+    return ths3
+
+----------------------------------------------------------------
+
+strategy :: EncodeStrategy
+strategy = EncodeStrategy { compressionAlgo = Linear, useHuffman = False }
+
+-- Set-Cookie: contains only one cookie value.
+-- So, we don't need to split it.
+hpackEncodeHeader :: Context -> Buffer -> BufSize
+                  -> TokenHeaderList
+                  -> IO (TokenHeaderList, Int)
+hpackEncodeHeader Context{..} buf siz ths =
+    encodeTokenHeader buf siz strategy True encodeDynamicTable ths
+
+hpackEncodeHeaderLoop :: Context -> Buffer -> BufSize
+                      -> TokenHeaderList
                       -> IO (TokenHeaderList, Int)
 hpackEncodeHeaderLoop Context{..} buf siz hs =
     encodeTokenHeader buf siz strategy False encodeDynamicTable hs
